use crate::prelude::*;

use asset::*;

use utilities::prelude::cgmath::{vec3, Deg, InnerSpace, Matrix4};

use std::sync::{Arc, Mutex, MutexGuard};

use std::time::Duration;
use utilities::{
    impl_reprc,
    prelude::cgmath::{Vector3, Vector4},
};

use anyhow::Result;

const ZERO_DURATION: Duration = Duration::from_secs(0);

create_settings_section!(
    ParticleSystemSpecification,
    "System",
    {
        // time in s
        duration: PersistentDuration,

        // total particle count
        particle_count: usize,

        // up vector
        up_x: f32,
        up_y: f32,
        up_z: f32,

        // right vector
        right_x: f32,
        right_y: f32,
        right_z: f32,

        // emissive factor
        emissive_r: f32,
        emissive_g: f32,
        emissive_b: f32,

        random_start_factor: f32,
        up_arc: f32,
        vertical_arc: f32,

        ratio: f32,
    },
    Serialize, Deserialize,
);

create_settings_section!(
    ParticleSpecification,
    "Particles",
    {
        // time in s
        minimum_life_time: PersistentDuration,
        maximum_life_time: PersistentDuration,

        // size in m
        start_size: f32,
        end_size: f32,

        // size deviation in per cent
        size_deviation: f32,

        // rgb colors
        start_color: Color,
        end_color: Color,

        // velocity in m/s
        minimum_velocity: f32,
        maximum_velocity: f32,

        // radius in m
        spawn_radius: f32,
    },
    Serialize, Deserialize,
);

create_settings_container!(
    ParticleSystemInfo,
    {
        system: ParticleSystemSpecification,
        particles: ParticleSpecification,
    },
    Serialize, Deserialize,
);

impl_reprc!(
    #[derive(Debug)]
    struct ParticleInfoBuffer {
        #[assume_reprc]
        start_size: f32,
        #[assume_reprc]
        end_size: f32,
        #[assume_reprc]
        start_time: f32,
        #[assume_reprc]
        life_time: f32,

        #[assume_reprc]
        start_position: Vector3<f32>,
        #[assume_reprc]
        velocity: f32,

        #[assume_reprc]
        direction: Vector3<f32>,
        #[assume_reprc]
        ratio: f32,

        #[assume_reprc]
        right: Vector3<f32>,
        #[assume_reprc]
        padding_1: f32,
    }
);

impl_reprc!(
    #[derive(Debug)]
    struct ParticleSystemInfoBuffer {
        #[assume_reprc]
        start_color: Vector4<f32>,
        #[assume_reprc]
        end_color: Vector4<f32>,
    }
);

impl ParticleSystemInfoBuffer {
    fn new(start_color: Vector3<f32>, end_color: Vector3<f32>) -> Self {
        Self {
            start_color: start_color.extend(1.0),
            end_color: end_color.extend(1.0),
        }
    }
}

pub struct ParticleSystemVulkanObjects {
    descriptor_set_layout: Arc<DescriptorSetLayout>,
    pipeline: Arc<Pipeline>,

    command_buffer: Arc<CommandBuffer>,

    recorder: Option<Mutex<CommandBufferRecorder<'static>>>,
}

unsafe impl Send for ParticleSystemVulkanObjects {}

unsafe impl Sync for ParticleSystemVulkanObjects {}

impl ParticleSystemVulkanObjects {
    pub(crate) fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        compute_shader: Arc<ShaderModule<shader_type::Compute>>,
    ) -> Result<Self> {
        let desc_layout = DescriptorSetLayout::builder()
            // particle system buffer
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            // particle buffer
            .add_layout_binding(
                1,
                VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            // vertex buffer
            .add_layout_binding(
                2,
                VK_DESCRIPTOR_TYPE_STORAGE_BUFFER,
                VK_SHADER_STAGE_COMPUTE_BIT,
                0,
            )
            .build(device.clone())?;

        let pipeline_layout = PipelineLayout::builder()
            .add_descriptor_set_layout(&desc_layout)
            .add_push_constant(VkPushConstantRange::new(VK_SHADER_STAGE_COMPUTE_BIT, 0, 4))
            .build(device.clone())?;

        let pipeline = Pipeline::new_compute()
            .set_shader_module(&compute_shader)
            .build(device, &pipeline_layout)?;

        Ok(Self {
            descriptor_set_layout: desc_layout,
            pipeline,

            command_buffer: CommandBuffer::new_secondary().build(device.clone(), queue.clone())?,

            recorder: Default::default(),
        })
    }

    pub(crate) fn begin(
        &mut self,
        begin_info: VkCommandBufferBeginInfo,
    ) -> Result<MutexGuard<'_, CommandBufferRecorder<'static>>> {
        if self.recorder.is_none() {
            let b = self.command_buffer.begin(begin_info)?;

            self.recorder = Some(Mutex::new(Self::rem_lt(b)));
        }

        Ok(self.recorder.as_ref().unwrap().lock().unwrap())
    }

    pub(crate) fn execute(
        &mut self,
        master_recorder: &mut CommandBufferRecorder<'_>,
    ) -> Result<()> {
        if let Some(recorder) = self.recorder.take() {
            // dropping recorder will end command buffer recording
            drop(recorder);

            master_recorder.execute_commands(&[&self.command_buffer]);
        }

        Ok(())
    }

    fn rem_lt<'a>(recorder: CommandBufferRecorder<'a>) -> CommandBufferRecorder<'static> {
        unsafe { std::mem::transmute(recorder) }
    }
}

pub struct ParticleSystem {
    info: ParticleSystemInfo,

    start: Duration,
    duration: Duration,

    random_start_factor: f32,

    _system_buffer: Arc<Buffer<ParticleSystemInfoBuffer>>,

    particle_buffer: Arc<Buffer<ParticleInfoBuffer>>,

    vertex_buffer: Arc<Buffer<PositionNormalColor>>,

    descriptor_set: Arc<DescriptorSet>,
}

impl ParticleSystem {
    pub fn new(
        info: ParticleSystemInfo,
        engine: &Engine,
        scene: &impl SceneEntities,
        draw: &mut Draw,
    ) -> Result<Self> {
        Self::_new(
            info,
            engine.device(),
            engine.queue(),
            scene.particle_system_vulkan_objects(),
            engine.settings().graphics_info()?.render_type,
            draw,
        )
    }

    pub fn new_from_scene<'a>(
        info: ParticleSystemInfo,
        scene: &Scene,
        draw: &mut Draw,
    ) -> Result<Self> {
        Self::_new(
            info,
            scene.device(),
            scene.queue(),
            scene.particle_system_vulkan_objects(),
            scene.render_type,
            draw,
        )
    }

    pub fn new_with_vk_objects(
        info: ParticleSystemInfo,
        engine: &Engine,
        particle_system_vulkan_objects: &ParticleSystemVulkanObjects,
        draw: &mut Draw,
    ) -> Result<Self> {
        Self::_new(
            info,
            engine.device(),
            engine.queue(),
            particle_system_vulkan_objects,
            engine.settings().graphics_info()?.render_type,
            draw,
        )
    }

    fn _new<'a>(
        info: ParticleSystemInfo,
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        particle_system_vulkan_objects: &ParticleSystemVulkanObjects,
        render_type: SceneType,
        draw: &'a mut Draw,
    ) -> Result<Self> {
        let (system_buffer, particle_buffer, vertex_buffer, desc_set) = {
            // create buffer directly on the gpu
            let command_buffer =
                CommandBuffer::new_primary().build(device.clone(), queue.clone())?;

            let (system_buffer, particle_buffer) = {
                let mut recorder = command_buffer.begin(VkCommandBufferBeginInfo::new(
                    VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT,
                ))?;

                let system_buffer =
                    Self::create_system_buffer(device, &mut recorder, &info.particles)?;

                let particle_buffer =
                    Self::create_particles(device, &mut recorder, &info.system, &info.particles)?;

                (system_buffer, particle_buffer)
            };

            // submit
            let submit = SubmitInfo::default().add_command_buffer(&command_buffer);
            let fence = Fence::builder().build(device.clone())?;

            queue.lock().unwrap().submit(Some(&fence), &[submit])?;

            device.wait_for_fences(&[&fence], true, Duration::from_secs(1))?;

            let mut mesh = AssetMesh::new(device, render_type)?;
            mesh.set_name("Particle System");

            let vertex_buffer: Arc<Buffer<PositionNormalColor>> = Buffer::builder()
                .set_memory_usage(MemoryUsage::GpuOnly)
                .set_usage(VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_VERTEX_BUFFER_BIT)
                .set_size(6 * info.system.particle_count.max(1) as VkDeviceSize)
                .build(device.clone())?;

            let material = PrimitiveMaterial {
                emissive_factor: [
                    info.system.emissive_r,
                    info.system.emissive_g,
                    info.system.emissive_b,
                ],

                ..Default::default()
            };

            mesh.add_primitive(vertex_buffer.clone(), None, None, material, true)?;

            let desc_set =
                Self::descriptor_set(particle_system_vulkan_objects.descriptor_set_layout.clone())?;

            desc_set.update(&[
                DescriptorWrite::storage_buffers(0, &[&system_buffer]),
                DescriptorWrite::storage_buffers(1, &[&particle_buffer]),
                DescriptorWrite::storage_buffers(2, &[&vertex_buffer]),
            ])?;

            draw.push(mesh);

            (system_buffer, particle_buffer, vertex_buffer, desc_set)
        };

        Ok(Self {
            start: Duration::from_secs(0),
            duration: info.system.duration.into(),

            random_start_factor: info.system.random_start_factor,

            info,

            _system_buffer: system_buffer,

            particle_buffer,

            vertex_buffer,

            descriptor_set: desc_set,
        })
    }

    pub fn info(&self) -> &ParticleSystemInfo {
        &self.info
    }

    fn descriptor_set(descriptor_layout: Arc<DescriptorSetLayout>) -> Result<Arc<DescriptorSet>> {
        let device = descriptor_layout.device().clone();

        DescriptorPool::builder()
            .set_layout(descriptor_layout)
            .build(device)?
            .prepare_set()
            .allocate()
    }

    /// `true` means that this component can live

    pub(crate) fn update(&self, time: Duration) -> bool {
        if self.duration != ZERO_DURATION && (self.start + self.duration) < time {
            return false;
        }

        true
    }

    pub(crate) fn animate(
        &self,
        recorder: &mut CommandBufferRecorder<'_>,
        scene: &SceneContents<'_>,
        time: Duration,
    ) -> Result<()> {
        recorder.bind_pipeline(&scene.particle_system_vulkan_objects().pipeline)?;
        recorder.bind_descriptor_sets_minimal(&[&self.descriptor_set]);
        recorder.push_constants(VK_SHADER_STAGE_COMPUTE_BIT, &time.as_secs_f32());
        recorder.dispatch(self.particle_buffer.size() as u32, 1, 1);

        recorder.buffer_barrier(
            &self.vertex_buffer,
            VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
            VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT,
            VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
        );

        Ok(())
    }

    fn create_system_buffer(
        device: &Arc<Device>,
        recorder: &mut CommandBufferRecorder<'_>,
        particle_info: &ParticleSpecification,
    ) -> Result<Arc<Buffer<ParticleSystemInfoBuffer>>> {
        Buffer::builder()
            .set_memory_usage(MemoryUsage::CpuToGpu)
            .set_usage(VK_BUFFER_USAGE_TRANSFER_SRC_BIT)
            .set_data(&[ParticleSystemInfoBuffer::new(
                Vector3::from({
                    let tmp: [f32; 3] = particle_info.start_color.into();
                    tmp
                }),
                Vector3::from({
                    let tmp: [f32; 3] = particle_info.end_color.into();
                    tmp
                }),
            )])
            .build(device.clone())?
            .into_device_local(
                recorder,
                VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT,
                VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
                VK_BUFFER_USAGE_STORAGE_BUFFER_BIT,
            )
    }

    fn create_particles(
        device: &Arc<Device>,
        recorder: &mut CommandBufferRecorder<'_>,
        spec: &ParticleSystemSpecification,
        particle_info: &ParticleSpecification,
    ) -> Result<Arc<Buffer<ParticleInfoBuffer>>> {
        let mut data = Vec::new();

        let up = vec3(spec.up_x, spec.up_y, spec.up_z).normalize();
        let right = vec3(spec.right_x, spec.right_y, spec.right_z).normalize();
        let mut normal = right.cross(up).normalize();

        if normal.z < 0.0 {
            normal = -normal;
        }

        let up_arc = spec.up_arc.abs();
        let vertical_arc = spec.vertical_arc.abs();

        for _ in 0..spec.particle_count.max(1) {
            let start_position = Random::range_f32(0.0, particle_info.spawn_radius)
                * (Matrix4::from_axis_angle(up, Deg(Random::range_f32(0.0, 360.0)))
                    * right.extend(1.0));

            let deviation = if particle_info.size_deviation == 0.0 {
                0.0
            } else {
                Random::range_f32(-particle_info.size_deviation, particle_info.size_deviation)
            };

            let minimum_life_time: Duration = particle_info.minimum_life_time.into();
            let maximum_life_time: Duration = particle_info.maximum_life_time.into();

            let (direction, right) = if up_arc == 0.0 {
                (up, right)
            } else {
                let up =
                    (Matrix4::from_axis_angle(normal, Deg(Random::range_f32(-up_arc, up_arc)))
                        * Matrix4::from_axis_angle(
                            right,
                            Deg(Random::range_f32(-vertical_arc, vertical_arc)),
                        )
                        * up.extend(1.0))
                    .truncate()
                    .normalize();

                let right = (-normal).cross(up).normalize();

                (up, right)
            };

            data.push(ParticleInfoBuffer {
                start_size: particle_info.start_size * (1.0 + deviation),
                end_size: particle_info.end_size * (1.0 + deviation),
                start_time: 0.0,
                life_time: if minimum_life_time.as_secs_f32() >= maximum_life_time.as_secs_f32() {
                    minimum_life_time.as_secs_f32()
                } else {
                    Random::range_f32(
                        minimum_life_time.as_secs_f32(),
                        maximum_life_time.as_secs_f32(),
                    )
                },
                velocity: if particle_info.minimum_velocity >= particle_info.maximum_velocity {
                    particle_info.minimum_velocity
                } else {
                    Random::range_f32(
                        particle_info.minimum_velocity,
                        particle_info.maximum_velocity,
                    )
                },
                start_position: start_position.truncate(),

                direction,
                right,

                ratio: if spec.ratio == 0.0 { 1.0 } else { spec.ratio },

                padding_1: 0.0,
            });
        }

        Buffer::builder()
            .set_memory_usage(MemoryUsage::CpuToGpu)
            .set_usage(VK_BUFFER_USAGE_TRANSFER_SRC_BIT)
            .set_data(&data)
            .build(device.clone())?
            .into_device_local(
                recorder,
                VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT,
                VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
                VK_BUFFER_USAGE_TRANSFER_SRC_BIT
                    | VK_BUFFER_USAGE_TRANSFER_DST_BIT
                    | VK_BUFFER_USAGE_STORAGE_BUFFER_BIT,
            )
    }
}

impl ParticleSystem {
    pub fn start_time(&self) -> Duration {
        self.start
    }

    pub fn duration(&self) -> Duration {
        self.duration
    }
}

impl EntityComponent for ParticleSystem {
    fn enable(&mut self, scene: &mut Scene) -> Result<()> {
        let now = scene.now();
        self.start = now;

        // Copy gpu only buffer into temporary cpu buffer to modify starting time
        // of each particle. Then copy the data back into the gpu only buffer.
        {
            let now_f32 = now.as_secs_f32();

            let fence = Fence::builder().build(scene.device().clone())?;
            let command_buffer = CommandBuffer::new_primary()
                .build(scene.device().clone(), scene.queue().clone())?;

            let cpu_buffer: Arc<Buffer<ParticleInfoBuffer>> = Buffer::builder()
                .set_memory_usage(MemoryUsage::GpuToCpu)
                .set_usage(VK_BUFFER_USAGE_TRANSFER_DST_BIT | VK_BUFFER_USAGE_TRANSFER_SRC_BIT)
                .set_size(self.particle_buffer.size())
                .build(scene.device().clone())?;

            {
                let mut recorder = command_buffer.begin(VkCommandBufferBeginInfo::new(
                    VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT,
                ))?;

                recorder.copy_buffer(
                    &self.particle_buffer,
                    &cpu_buffer,
                    &[VkBufferCopy {
                        srcOffset: 0,
                        dstOffset: 0,
                        size: self.particle_buffer.byte_size(),
                    }],
                );
            }

            // submit
            let submit = SubmitInfo::default().add_command_buffer(&command_buffer);
            scene
                .queue()
                .lock()
                .unwrap()
                .submit(Some(&fence), &[submit])?;
            scene
                .device()
                .wait_for_fences(&[&fence], true, Duration::from_secs(1))?;

            fence.reset();

            // apply now as starting point
            {
                let mut mapped = cpu_buffer.map_complete()?;

                for particle in mapped.iter_mut() {
                    particle.start_time = now_f32;

                    if self.random_start_factor != 0.0 {
                        particle.start_time -=
                            Random::range_f32(0.0, particle.life_time * self.random_start_factor);
                    }
                }
            }

            {
                let mut recorder = command_buffer.begin(VkCommandBufferBeginInfo::new(
                    VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT,
                ))?;

                recorder.copy_buffer(
                    &cpu_buffer,
                    &self.particle_buffer,
                    &[VkBufferCopy {
                        srcOffset: 0,
                        dstOffset: 0,
                        size: self.particle_buffer.byte_size(),
                    }],
                );

                recorder.buffer_barrier(
                    &self.particle_buffer,
                    VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT,
                    VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
                    VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT,
                    VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
                );
            }

            // submit
            let submit = SubmitInfo::default().add_command_buffer(&command_buffer);
            scene
                .queue()
                .lock()
                .unwrap()
                .submit(Some(&fence), &[submit])?;
            scene
                .device()
                .wait_for_fences(&[&fence], true, Duration::from_secs(1))?;
        }

        Ok(())
    }

    fn name(&self) -> &str {
        "ParticleSystem"
    }
}

impl ComponentDebug for ParticleSystem {
    fn debug_name() -> &'static str {
        "ParticleSystem"
    }
}

impl ComponentCreateInfo<'_> for ParticleSystemInfo {}
