use std::{
    any::{Any, TypeId},
    collections::HashMap,
    ptr::NonNull,
    time::{Duration, Instant},
};

use utilities::prelude::cgmath::Vector3;

use anyhow::{bail, Error, Result};
use indexmap::IndexMap;

use crate::scene::{content::events::Events, rendering::RenderingFrontEnd};

use crate::prelude::*;

use super::scene_base::EntityMultiMut;

#[derive(Debug)]
pub struct EntityNotFoundError {
    entity: Entity,
}

impl EntityNotFoundError {
    pub(crate) fn new(entity: Entity) -> Self {
        Self { entity }
    }
}

impl std::fmt::Display for EntityNotFoundError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Entity (ID: {}) not found!", self.entity)
    }
}

impl std::error::Error for EntityNotFoundError {}

pub trait SceneEntities {
    fn entity(&self, entity: Entity) -> std::result::Result<&EntityObject, EntityNotFoundError>;
    fn entity_mut(
        &mut self,
        entity: Entity,
    ) -> std::result::Result<&mut EntityObject, EntityNotFoundError>;
    fn entities_multi_mut(&mut self) -> EntityMultiMut<'_>;
    fn add_entity(&mut self, entity_object: EntityObject) -> Result<Entity>;
    fn archetype_info(&self, name: impl ToString) -> Result<ArchetypeInfo>;

    fn insert_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
        component: T,
    ) -> Result<()>;

    fn remove_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
    ) -> Result<()>;

    fn write_event<T: Any + Send + Sync>(&mut self, payload: T);

    // This function will panic if map is not set!
    fn map<F, T>(&mut self, f: F) -> Result<T>
    where
        F: FnOnce(&Box<dyn Map>, &mut SceneContents<'_>) -> Result<T>;

    fn remove_entity(&mut self, entity: Entity) -> Result<Option<EntityObject>>;
    fn now(&self) -> Duration;

    fn particle_system_vulkan_objects_mut(&mut self) -> &mut ParticleSystemVulkanObjects;

    fn particle_system_vulkan_objects(&self) -> &ParticleSystemVulkanObjects;

    fn world_to_screen_space(&self, world_space: Vector3<f32>) -> Result<(i32, i32)>;

    fn screen_space_to_world(&self, x: u32, y: u32) -> Result<Option<Vector3<f32>>>;

    fn add_light(&mut self, light: Light) -> Result<()>;
    fn remove_light(&mut self, light: Light) -> Result<()>;
    fn clear_lights(&mut self) -> Result<()>;

    fn view_mut(&mut self) -> &mut View;

    fn render_type(&self) -> SceneType;
    fn frame_time(&self) -> Duration;
}

enum ComponentChange {
    Added(TypeId, Box<dyn EntityComponent>),
    Removed(TypeId),
}

pub struct SceneContents<'a> {
    render_type: SceneType,
    frame_time: Duration,

    screen_extents: (f32, f32),

    start_time: Instant,

    map: &'a Option<Box<dyn Map>>,

    renderer: &'a mut Box<dyn RenderingFrontEnd + Send + Sync>,

    particle_system_vulkan_objects: &'a mut ParticleSystemVulkanObjects,

    archetypes: &'a Archetypes,
    pub events: ContentEvents<'a>,
    pub entities: Entities<'a>,
    pub resources: &'a mut Resources,
}

impl<'a> SceneContents<'a> {
    pub(crate) fn new(
        render_type: SceneType,
        frame_time: Duration,
        start_time: Instant,
        screen_extents: (f32, f32),
        entities: &'a mut IndexMap<Entity, EntityObject>,
        map: &'a Option<Box<dyn Map>>,
        (renderer, particle_system_vulkan_objects): (
            &'a mut Box<dyn RenderingFrontEnd + Send + Sync>,
            &'a mut ParticleSystemVulkanObjects,
        ),
        archetypes: &'a Archetypes,
        events: &'a mut Events,
        resources: &'a mut Resources,
    ) -> Self {
        Self {
            screen_extents,
            frame_time,

            start_time,
            map,

            renderer,

            particle_system_vulkan_objects,

            archetypes,
            events: ContentEvents { events },
            entities: Entities {
                entities,
                entities_to_remove: Vec::new(),
                entities_to_add: Vec::new(),
                entities_updates: HashMap::new(),
            },
            resources,

            render_type,
        }
    }

    pub fn entity_changes(self) -> EntityChanges {
        EntityChanges {
            entities_to_remove: self.entities.entities_to_remove,
            entities_to_add: self.entities.entities_to_add,
            entities_updates: self.entities.entities_updates,
        }
    }
}

pub struct ContentEvents<'a> {
    pub(crate) events: &'a mut Events,
}

impl<'a> ContentEvents<'a> {
    pub fn write_event<T: Any + Send + Sync>(&mut self, payload: T) {
        self.events.write_event(payload);
    }
}

pub struct Entities<'a> {
    pub(crate) entities: &'a mut IndexMap<Entity, EntityObject>,
    entities_to_remove: Vec<Entity>,
    entities_to_add: Vec<EntityObject>,
    entities_updates: HashMap<Entity, Vec<ComponentChange>>,
}

impl<'a> Entities<'a> {
    pub fn entity(
        &self,
        entity: Entity,
    ) -> std::result::Result<&EntityObject, EntityNotFoundError> {
        self.entities
            .get(&entity)
            .ok_or_else(|| EntityNotFoundError::new(entity))
    }

    pub unsafe fn entity_unchecked(
        &self,
        entity: Entity,
    ) -> std::result::Result<&'a EntityObject, EntityNotFoundError> {
        self.entity(entity).map(|e| remove_life_time(e))
    }

    pub fn entity_mut(
        &mut self,
        entity: Entity,
    ) -> std::result::Result<&mut EntityObject, EntityNotFoundError> {
        self.entities
            .get_mut(&entity)
            .ok_or_else(|| EntityNotFoundError::new(entity))
    }

    pub unsafe fn entity_mut_unchecked(
        &mut self,
        entity: Entity,
    ) -> std::result::Result<&'a mut EntityObject, EntityNotFoundError> {
        self.entity_mut(entity).map(|e| remove_life_time_mut(e))
    }

    pub fn add_entity(&mut self, entity_object: EntityObject) -> Result<Entity> {
        let entity = entity_object.as_entity();
        self.entities_to_add.push(entity_object);

        Ok(entity)
    }

    pub fn insert_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
        component: T,
    ) -> Result<()> {
        println!("insert component {}", T::debug_name());

        let change = ComponentChange::Added(TypeId::of::<T>(), Box::new(component));

        match self.entities_updates.get_mut(&entity) {
            Some(changes) => {
                changes.push(change);
            }
            None => {
                self.entities_updates.insert(entity, vec![change]);
            }
        }

        Ok(())
    }

    pub fn remove_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
    ) -> Result<()> {
        println!("remove component {}", T::debug_name());

        let change = ComponentChange::Removed(TypeId::of::<T>());

        match self.entities_updates.get_mut(&entity) {
            Some(changes) => {
                changes.push(change);
            }
            None => {
                self.entities_updates.insert(entity, vec![change]);
            }
        }

        Ok(())
    }

    pub fn remove_entity(&mut self, entity: Entity) -> Result<Option<EntityObject>> {
        self.entities_to_remove.push(entity);

        Ok(None)
    }

    pub fn entities_multi_mut(&mut self) -> EntityMultiMut<'_> {
        EntityMultiMut::new(self.entities)
    }
}

pub struct EntityChanges {
    entities_to_remove: Vec<Entity>,
    entities_to_add: Vec<EntityObject>,
    entities_updates: HashMap<Entity, Vec<ComponentChange>>,
}

impl EntityChanges {
    pub fn commit(self, scene: &mut Scene) -> Result<()> {
        if !self.entities_to_remove.is_empty() {
            println!("entities to remove {}", self.entities_to_remove.len());
        }

        for entity in self.entities_to_remove {
            scene.remove_entity(entity)?;
        }

        if !self.entities_to_add.is_empty() {
            println!("entities to add {}", self.entities_to_add.len());
        }

        for entity_object in self.entities_to_add {
            scene.add_entity(entity_object)?;
        }

        if !self.entities_updates.is_empty() {
            println!("entities to update {}", self.entities_updates.len());
        }

        for (entity, changes) in self.entities_updates {
            scene.updates.remove_entity(entity);
            scene.archetypes.remove_entity(entity);

            if let Some(entity_object) = unsafe {
                scene
                    .entities
                    .get_mut(&entity)
                    .map(|e| remove_life_time_mut(e))
            } {
                entity_object.activation_state.apply_change();

                for change in changes {
                    match change {
                        ComponentChange::Added(type_id, mut component) => {
                            component.enable(scene)?;
                            let name = component.name().to_string();

                            if entity_object
                                .insert_component_by_id(type_id, component)
                                .is_some()
                            {
                                bail!("component {name} already present");
                            }
                        }
                        ComponentChange::Removed(type_id) => {
                            if let Some(mut component) =
                                entity_object.remove_component_by_id(type_id)
                            {
                                component.disable(scene)?;
                            }
                        }
                    }
                }

                entity_object.activation_state.apply_change();

                scene.updates.add_entity(entity_object, &scene.entities)?;
                scene.archetypes.add_entity(entity_object)?;
            }
        }

        Ok(())
    }
}

impl<'a> SceneEntities for SceneContents<'a> {
    fn entity(&self, entity: Entity) -> std::result::Result<&EntityObject, EntityNotFoundError> {
        self.entities.entity(entity)
    }

    fn entity_mut(
        &mut self,
        entity: Entity,
    ) -> std::result::Result<&mut EntityObject, EntityNotFoundError> {
        self.entities.entity_mut(entity)
    }

    fn add_entity(&mut self, entity_object: EntityObject) -> Result<Entity> {
        self.entities.add_entity(entity_object)
    }

    fn archetype_info(&self, name: impl ToString) -> Result<ArchetypeInfo> {
        let name = name.to_string();
        let archetype = self
            .archetypes
            .get(&name)
            .ok_or_else(|| Error::msg(format!("requested archetype ({}) not found", name)))?;

        let mut entities = Vec::new();

        for entity in archetype.entities().keys() {
            #[cfg(debug_assertions)]
            let debug_name = self.entity(*entity)?.debug_name.clone();

            #[cfg(not(debug_assertions))]
            let debug_name = None;

            entities.push((*entity, debug_name));
        }

        Ok(ArchetypeInfo::new(entities))
    }

    fn insert_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
        component: T,
    ) -> Result<()> {
        self.entities.insert_component(entity, component)
    }

    fn remove_component<T: EntityComponent + ComponentDebug>(
        &mut self,
        entity: Entity,
    ) -> Result<()> {
        self.entities.remove_component::<T>(entity)
    }

    fn remove_entity(&mut self, entity: Entity) -> Result<Option<EntityObject>> {
        self.entities.remove_entity(entity)
    }

    fn entities_multi_mut(&mut self) -> EntityMultiMut<'_> {
        self.entities.entities_multi_mut()
    }

    fn write_event<T: Any + Send + Sync>(&mut self, payload: T) {
        self.events.write_event(payload);
    }

    // This function will panic if map is not set!
    fn map<F, T>(&mut self, f: F) -> Result<T>
    where
        F: FnOnce(&Box<dyn Map>, &mut SceneContents<'a>) -> Result<T>,
    {
        let map = self
            .map
            .as_ref()
            .expect("Map is not set while trying to call map");
        let p = map as *const Box<dyn Map> as *mut Box<dyn Map>;

        f(unsafe { NonNull::new_unchecked(p).as_ref() }, self)
    }

    fn now(&self) -> Duration {
        Instant::now().duration_since(self.start_time)
    }

    fn world_to_screen_space(&self, world_space: Vector3<f32>) -> Result<(i32, i32)> {
        Scene::_world_to_screen_space(self.screen_extents, world_space, self.renderer.view())
    }

    fn screen_space_to_world(&self, x: u32, y: u32) -> Result<Option<Vector3<f32>>> {
        let scale = self.renderer.render_scale();

        self.renderer
            .screen_to_world((x as f32 * scale) as u32, (y as f32 * scale) as u32)
    }

    fn particle_system_vulkan_objects_mut(&mut self) -> &mut ParticleSystemVulkanObjects {
        &mut self.particle_system_vulkan_objects
    }

    fn particle_system_vulkan_objects(&self) -> &ParticleSystemVulkanObjects {
        self.particle_system_vulkan_objects
    }

    fn add_light(&mut self, light: Light) -> Result<()> {
        self.renderer.add_light(light)?;

        Ok(())
    }

    fn remove_light(&mut self, light: Light) -> Result<()> {
        self.renderer.remove_light(light)?;

        Ok(())
    }

    fn clear_lights(&mut self) -> Result<()> {
        self.renderer.clear_lights()?;

        Ok(())
    }

    fn view_mut(&mut self) -> &mut View {
        self.renderer.view_mut()
    }

    fn render_type(&self) -> SceneType {
        self.render_type
    }

    fn frame_time(&self) -> Duration {
        self.frame_time
    }
}
