use anyhow::bail;
use engine::prelude::{image::RgbaImage, *};

use std::{cmp::Ordering, slice::Iter, str::FromStr};

use crate::{
    config::attributes::{AttributeColorSettings, StartingAttributes},
    items::{ability_book::Ability, ItemSystem},
};

generate_stat!(Agility, u32, "Agility");
generate_stat!(Intelligence, u32, "Intelligence");
generate_stat!(Strength, u32, "Strength");

#[derive(Debug, Clone, Copy, Hash, PartialEq, Eq, PartialOrd)]
pub enum Attribute {
    Agility,
    Intelligence,
    Strength,
}

impl Attribute {
    pub fn random() -> Self {
        match Random::range(0, 3) {
            0 => Self::Agility,
            1 => Self::Intelligence,
            2 => Self::Strength,

            _ => unreachable!(),
        }
    }

    pub fn iter() -> Iter<'static, Self> {
        use Attribute::*;

        static ATTRIBUTES: [Attribute; 3] = [Agility, Intelligence, Strength];
        ATTRIBUTES.iter()
    }

    pub fn apply_color<A: Ability>(
        &self,
        base_image: &RgbaImage,
        color_settins: &AttributeColorSettings,
    ) -> RgbaImage {
        ItemSystem::<A>::apply_color(base_image, color_settins.from_attribute(*self))
    }
}

impl FromStr for Attribute {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "agility" => Ok(Self::Agility),
            "intelligence" => Ok(Self::Intelligence),
            "strength" => Ok(Self::Strength),

            _ => bail!("could not parse attribute from {s}"),
        }
    }
}

impl std::fmt::Display for Attribute {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Attribute::Agility => write!(f, "agility"),
            Attribute::Intelligence => write!(f, "intelligence"),
            Attribute::Strength => write!(f, "strength"),
        }
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Attributes {
    agility: Agility,
    intelligence: Intelligence,
    strength: Strength,

    pub bonus_agility: Agility,
    pub bonus_intelligence: Intelligence,
    pub bonus_strength: Strength,
}

impl Attributes {
    pub fn zero() -> Self {
        Attributes {
            agility: Agility::default(),
            intelligence: Intelligence::default(),
            strength: Strength::default(),

            bonus_agility: Agility::default(),
            bonus_intelligence: Intelligence::default(),
            bonus_strength: Strength::default(),
        }
    }

    pub fn new(starting_attributes: &StartingAttributes) -> Self {
        Attributes {
            agility: Agility::from(starting_attributes.agility),
            intelligence: Intelligence::from(starting_attributes.intelligence),
            strength: Strength::from(starting_attributes.strength),

            bonus_agility: Agility::default(),
            bonus_intelligence: Intelligence::default(),
            bonus_strength: Strength::default(),
        }
    }

    pub fn load(strength: u32, agility: u32, intelligence: u32) -> Self {
        Attributes {
            agility: Agility::from(agility),
            intelligence: Intelligence::from(intelligence),
            strength: Strength::from(strength),

            bonus_agility: Agility::default(),
            bonus_intelligence: Intelligence::default(),
            bonus_strength: Strength::default(),
        }
    }

    pub fn apply_boni(&mut self, strength: Strength, agility: Agility, intelligence: Intelligence) {
        self.strength = strength;
        self.agility = agility;
        self.intelligence = intelligence;
    }

    pub fn random_with_base(sum: u32, starting_attributes: &StartingAttributes) -> Self {
        let me = Self::new(starting_attributes);
        let mut random = Self::zero();
        random.randomize(sum);

        me + random
    }

    pub fn agility(&self) -> Agility {
        self.agility + self.bonus_agility
    }

    pub fn strength(&self) -> Strength {
        self.strength + self.bonus_strength
    }

    pub fn intelligence(&self) -> Intelligence {
        self.intelligence + self.bonus_intelligence
    }

    pub fn base_agility(&self) -> Agility {
        self.agility
    }

    pub fn base_strength(&self) -> Strength {
        self.strength
    }

    pub fn base_intelligence(&self) -> Intelligence {
        self.intelligence
    }

    pub fn add_agility(&mut self, agility: Agility) {
        self.agility += agility;
    }

    pub fn add_strength(&mut self, strength: Strength) {
        self.strength += strength;
    }

    pub fn add_intelligence(&mut self, intelligence: Intelligence) {
        self.intelligence += intelligence;
    }

    pub fn base_sum(&self) -> u32 {
        self.agility.raw() + self.intelligence.raw() + self.strength.raw()
    }
}

impl Attributes {
    pub fn randomize(&mut self, sum: u32) {
        enum AttributeTypes {
            Strength,
            Agility,
            Intelligence,
        }

        let mut attrs = vec![
            AttributeTypes::Strength,
            AttributeTypes::Agility,
            AttributeTypes::Intelligence,
        ];

        let n = Random::range(0, 3) as usize;
        let attr_type = attrs.remove(n);

        let val1 = match attr_type {
            AttributeTypes::Strength => {
                self.strength.set(Random::range(0, sum + 1));
                self.strength.raw()
            }
            AttributeTypes::Agility => {
                self.agility.set(Random::range(0, sum + 1));
                self.agility.raw()
            }
            AttributeTypes::Intelligence => {
                self.intelligence.set(Random::range(0, sum + 1));
                self.intelligence.raw()
            }
        };

        let n = Random::range(0, 2) as usize;
        let attr_type = attrs.remove(n);

        let val2 = match attr_type {
            AttributeTypes::Strength => {
                self.strength.set(Random::range(0, (sum - val1) + 1));
                self.strength.raw()
            }
            AttributeTypes::Agility => {
                self.agility.set(Random::range(0, (sum - val1) + 1));
                self.agility.raw()
            }
            AttributeTypes::Intelligence => {
                self.intelligence.set(Random::range(0, (sum - val1) + 1));
                self.intelligence.raw()
            }
        };

        let attr_type = attrs.remove(0);

        match attr_type {
            AttributeTypes::Strength => {
                self.strength.set(sum - (val1 + val2));
                self.strength.raw()
            }
            AttributeTypes::Agility => {
                self.agility.set(sum - (val1 + val2));
                self.agility.raw()
            }
            AttributeTypes::Intelligence => {
                self.intelligence.set(sum - (val1 + val2));
                self.intelligence.raw()
            }
        };

        debug_assert!(sum == self.sum());
    }

    pub fn sum(&self) -> u32 {
        self.agility.raw() + self.intelligence.raw() + self.strength.raw()
    }
}

impl std::ops::Add for Attributes {
    type Output = Self;

    fn add(self, other: Attributes) -> Self {
        Attributes {
            agility: self.agility + other.agility,
            strength: self.strength + other.strength,
            intelligence: self.intelligence + other.intelligence,

            bonus_agility: Agility::default(),
            bonus_intelligence: Intelligence::default(),
            bonus_strength: Strength::default(),
        }
    }
}

impl PartialEq for Attributes {
    fn eq(&self, other: &Self) -> bool {
        self.intelligence == other.intelligence
            && self.agility == other.agility
            && self.strength == other.strength
    }
}

impl Eq for Attributes {}

impl PartialOrd for Attributes {
    fn partial_cmp(&self, other: &Attributes) -> Option<Ordering> {
        let strength = self.strength.raw().cmp(&other.strength.raw());

        if strength == Ordering::Less {
            return Some(strength);
        }

        let agility = self.agility.raw().cmp(&other.agility.raw());

        if agility == Ordering::Less {
            return Some(agility);
        }

        let intelligence = self.intelligence.raw().cmp(&other.intelligence.raw());

        if intelligence == Ordering::Less {
            return Some(intelligence);
        }

        if strength == Ordering::Equal
            && agility == Ordering::Equal
            && intelligence == Ordering::Equal
        {
            return Some(Ordering::Equal);
        }

        if strength == Ordering::Greater
            && agility == Ordering::Greater
            && intelligence == Ordering::Greater
        {
            return Some(Ordering::Greater);
        }

        None
    }
}

impl EntityComponent for Attributes {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for Attributes {
    fn debug_name() -> &'static str {
        "Attributes"
    }
}
